import subprocess

from uceasy.adapters.uce_processor import UCEProcessor


class UCEPhylogenomics:


    def __init__(self, output, log, contigs, probes, taxon_set_conf, taxa, aligner, charsets, percent, internal_trimming):
        self.__processor = UCEProcessor()
        self.__output = output
        self.__contigs = contigs
        self.__probes = probes
        self.__log = log
        self.__taxon_set_conf = taxon_set_conf
        self.__taxa = taxa
        self.__aligner = aligner
        self.__charsets = charsets
        self.__percent = percent
        self.__internal_trimming = internal_trimming


    def run_uce_processing(self):

        self._get_match_contigs_to_probes()
        self._get_get_match_counts()
        self._get_get_fastas_from_match_counts()
        self._get_explode_get_fastas_file()
        self._get_secap_align()
        if self.__internal_trimming:
            self._get_get_gblocks_trimmed_alignments_from_untrimmed()
            self._get_remove_locus_name_from_nexus_lines()
            self._get_get_only_loci_with_min_taxa()
        self._get_nexus_files_for_raxml()

        
    def _get_match_contigs_to_probes(self):
        self.__locus_db = self.__output + '/uce_search_results/probe.matches.sqlite'

        return self.__processor.match_contigs_to_probes(self.__output + '/uce_search_results',
                                                        self.__contigs,
                                                        self.__probes)
       
    
    def _get_get_match_counts(self):
        self.__match_count_output = self.__output + '/all-taxa-incomplete.conf'
        self.__incomplete_matrix = self.__output + '/all-taxa-incomplete.incomplete'

        return self.__processor.get_match_counts(self.__match_count_output,
                                                 self.__locus_db,
                                                 self.__taxon_set_conf,
                                                 'all')
   

    def _get_get_fastas_from_match_counts(self):
        self.__alignments = self.__output + '/all-taxa-incomplete.fasta'

        return self.__processor.get_fastas_from_match_counts(self.__alignments,
                                                             self.__contigs,
                                                             self.__locus_db,
                                                             self.__match_count_output,
                                                             self.__incomplete_matrix,
                                                             self.__log)
    

    def _get_explode_get_fastas_file(self):
        return self.__processor.explode_get_fastas_file(self.__alignments, self.__output + '/exploded_fastas')


    def _get_secap_align(self):
        self.__nexus_files = self.__output + '/nexus_files'

        return self.__processor.seqcap_align(self.__nexus_files,
                                             self.__alignments,
                                             self.__taxa,
                                             self.__aligner,
                                             self.__internal_trimming)


    def _get_get_gblocks_trimmed_alignments_from_untrimmed(self):
        cmd = self.__processor.get_gblocks_trimmed_alignments_from_untrimmed(self.__nexus_files + '_gblocks',
                                                                             self.__nexus_files,
                                                                             self.__log)

        self.__nexus_files = self.__nexus_files + '_gblocks'
        return cmd


    def _get_remove_locus_name_from_nexus_lines(self):
        cmd = self.__processor.remove_locus_name_from_nexus_lines(self.__nexus_files + '_clean',
                                                                  self.__nexus_files,
                                                                   self.__log)
        self.__nexus_files = self.__nexus_files + '_clean'
        return cmd
    

    def _get_get_only_loci_with_min_taxa(self):
        cmd = self.__processor.get_only_loci_with_min_taxa(self.__nexus_files + '_min_taxa',
                                                            self.__nexus_files,
                                                            self.__taxa,
                                                            self.__percent,
                                                            self.__log)
        self.__nexus_files = self.__nexus_files + '_min_taxa'
        return cmd


    def _get_nexus_files_for_raxml(self):
        cmd = self.__processor.format_nexus_files_for_raxml(self.__nexus_files + '_raxml',
                                                             self.__nexus_files,
                                                             self.__log,
                                                             self.__charsets)
        self.__nexus_files = self.__nexus_files = '_raxml'
        return cmd



import subprocess
import os
import shutil
from uceasy.adapters import CPU, ILLUMIPROCESSOR


def run_illumiprocessor(input, output, config):

    # Prevent Illumiprocessor from asking to remove output dir
    if os.path.isdir(output):
        shutil.rmtree(output)

    cmd = [
        ILLUMIPROCESSOR,
        '--input', input,
        '--output', output,
        '--config', config,
        '--cores', CPU,
    ]
    return subprocess.run(cmd, check=True)



from multiprocessing import cpu_count
import json
import subprocess
import os


# Get environment variable UCEASY_PHYLUCE, default: phyluce
phyluce_env = os.getenv('UCEASY_PHYLUCE', 'phyluce')
cmd = ['conda', 'search', '--envs', '--json', phyluce_env]
out = subprocess.check_output(cmd)
phyluce_json = json.loads(out)


try:
    PHYLUCE = phyluce_json[0]['location']
except IndexError:
    raise IOError('Phyluce was not found, specify the conda environment with the UCEASY_PHYLUCE variable.')


TRIMMOMATIC     = PHYLUCE + '/bin/trimmomatic'
ILLUMIPROCESSOR = PHYLUCE + '/bin/illumiprocessor'
TRINITY         = PHYLUCE + '/bin/phyluce_assembly_assemblo_trinity'
SPADES          = PHYLUCE + '/bin/phyluce_assembly_assemblo_spades'

CPU = str(cpu_count() - 2)



from uceasy.adapters import CPU, TRINITY, SPADES
import subprocess


def run_trinity(config, output):
    cmd = [
        TRINITY,
        '--config', config,
        '--output', output,
        '--clean',
        '--cores', CPU
    ]
    return subprocess.run(cmd, check=True)


def run_spades(config, output):
    cmd = [
        SPADES,
        '--config', config,
        '--output', output,
        '--cores', CPU
    ]
    return subprocess.run(cmd, check=True)




from collections import namedtuple


params = (
        'input',
        'output',
        'sheet',
        'adapter_i7',
        'adapter_i5',
        'probes',
        'assembler',
        'aligner',
        'charsets',
        'percent',
        'internal_trimming',

        # OPTIONAL
        'fastq_encoding',
        'min_len',
        'r1_pattern',
        'r2_pattern',
        'no_merge',
        'single_end'
        )

defaults = (
        None,
        None,
        None,
        None,
        None,
        None
        )

Context = namedtuple('Context', params, defaults=defaults)




from uceasy.adapters import CPU, PHYLUCE
import subprocess


class UCEProcessor:


    def runner(self, cmd):
        return subprocess.run(cmd, check=True)


    def match_contigs_to_probes(self, output, contigs, probes):
        cmd = [
            PHYLUCE + '/bin/phyluce_assembly_match_contigs_to_probes',
            '--contigs', contigs,
            '--probes', probes,
            '--output', output
        ]
        return self.runner(cmd)


    def get_match_counts(self, output, locus_db, taxon_list_config, taxon_group):
        cmd = [
            PHYLUCE + '/bin/phyluce_assembly_get_match_counts',
            '--locus-db', locus_db,
            '--taxon-list-config', taxon_list_config,
            '--taxon-group', taxon_group,
            '--output', output,
            '--incomplete-matrix'
        ]
        return self.runner(cmd)


    def get_fastas_from_match_counts(self, output, contigs, locus_db, match_count_output, incomplete_matrix, log):
        cmd = [
            PHYLUCE + '/bin/phyluce_assembly_get_fastas_from_match_counts',
            '--contigs', contigs,
            '--locus-db', locus_db,
            '--match-count-output', match_count_output,
            '--output', output,
            '--log-path', log,
            '--incomplete-matrix', incomplete_matrix
        ]
        return self.runner(cmd)


    def explode_get_fastas_file(self, input, output):
        cmd = [
            PHYLUCE + '/bin/phyluce_assembly_explode_get_fastas_file',
            '--input', input,
            '--output', output,
            '--by-taxon'
        ]
        return self.runner(cmd)

    def seqcap_align(self, output, fasta, taxa, aligner, no_trim=False):
        cmd = [
            PHYLUCE + '/bin/phyluce_align_seqcap_align',
            '--fasta', fasta,
            '--taxa', taxa,
            '--aligner', aligner,
            '--output', output,
            '--output-format', 'fasta',
            '--incomplete-matrix'
        ]
        if no_trim:
            cmd.append('--no-trim')
        return self.runner(cmd)


    def get_gblocks_trimmed_alignments_from_untrimmed(self, output, alignments, log):
        cmd = [
            PHYLUCE + '/bin/phyluce_align_get_gblocks_trimmed_alignments_from_untrimmed',
            '--alignments', alignments,
            '--output', output,
            '--log', log
            ]
        return self.runner(cmd)


    def remove_locus_name_from_nexus_lines(self, output, alignments, log):
        cmd = [
            PHYLUCE + '/bin/phyluce_align_remove_locus_name_from_nexus_lines',
            '--alignments', alignments,
            '--output', output,
            '--log-path', log,
            '--cores', CPU
        ]
        return self.runner(cmd)


    def get_only_loci_with_min_taxa(self, output, alignments, taxa, percent, log):
        cmd = [
            PHYLUCE + '/bin/phyluce_align_get_only_loci_with_min_taxa',
            '--alignments', alignments,
            '--taxa', taxa,
            '--percent', percent,
            '--output', output,
            '--cores', CPU,
            '--log-path', log
        ]
        return self.runner(cmd)


    def format_nexus_files_for_raxml(self, output, alignments, log, charsets=False):
        cmd = [
            PHYLUCE + '/bin/phyluce_align_format_nexus_files_for_raxml',
            '--alignments', alignments,
            '--output', output,
            '--log-path', log
        ]
        if charsets:
            cmd.append('--charsets')
        return self.runner(cmd)




# -*- coding: utf-8 -*-
from distutils.core import setup

packages = \
['uceasy', 'uceasy.adapters', 'uceasy.controller', 'uceasy.use_cases']

package_data = \
{'': ['*']}

install_requires = \
['click>=7.0,<8.0', 'flask>=1.1,<2.0', 'pandas>=0.25.0,<0.26.0']

entry_points = \
{'console_scripts': ['uceasy = cli.uceasy_cli:uceasy']}

setup_kwargs = {
    'name': 'uceasy',
    'version': '0.1.1',
    'description': 'Wrapper for the Phyluce phylogenomics software package',
    'long_description': None,
    'author': 'Caio Raposo',
    'author_email': 'caioraposo@protonmail.com',
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



import os

from uceasy.adapters import assembly
from uceasy.adapters import quality_control
from uceasy.controller import env_manager
from uceasy.use_cases.uce_phylogenomics import UCEPhylogenomics


class Facade:


    def quality_control(self, input, output, sheet, adapter_i7, adapter_i5):

        config_dict = env_manager.prepare_illumiprocessor_conf(sheet,
                                                               adapter_i7,
                                                               adapter_i5)

        config = env_manager.render_conf_file(output + '/illumiprocessor.conf', config_dict)

        return quality_control.run_illumiprocessor(input, output + '/clean_fastq', config)


    def assembly(self, output, assembler, samples):

        config_dict = env_manager.prepare_assembly_conf(output, samples)
        config = env_manager.render_conf_file(output + '/assembly.conf', config_dict)

        return assembly.run_spades(config, output + '/assembly')


    def process_uce(self, output, log, contigs, probes, samples, aligner, charsets, percent, internal_trimming):
        taxon_set_conf = env_manager.render_taxon_set_conf(output + '/taxon-set.conf', samples)

        taxa = str(len(samples))

        processor = UCEPhylogenomics(output, log, contigs, probes, taxon_set_conf, taxa, aligner, charsets, percent, internal_trimming)
        processor.run_uce_processing()



from uceasy.controller.phyluce_facade import Facade
from uceasy.controller import env_manager


class Interactor:


    def __init__(self, context):
        self.__context = context
        self.__facade = Facade()


    def run(self):

        samples = env_manager.get_samples_from_csv(self.__context.sheet)

        self.__facade.quality_control(self.__context.input,
                                      self.__context.output,
                                      self.__context.sheet,
                                      self.__context.adapter_i7,
                                      self.__context.adapter_i5)

        self.__facade.assembly(self.__context.output, self.__context.assembler, samples)
        
        contigs = self.__context.output + '/assembly/contigs'

        self.__facade.process_uce(self.__context.output,
                                  self.__context.output,
                                  contigs,
                                  self.__context.probes,
                                  samples,
                                  self.__context.aligner,
                                  self.__context.charsets,
                                  self.__context.percent,
                                  self.__context.internal_trimming)




import pandas as pd
import os
import configparser


def render_conf_file(file, config_dict):
    config = configparser.ConfigParser(delimiters=(':'))
    config.optionxform = str
    config.read_dict(config_dict)

    with open(file, 'w') as fl:
        config.write(fl, space_around_delimiters=False)

    return file


def prepare_illumiprocessor_conf(sheet, adapter_i7, adapter_i5):
    config_dict = dict()
    sheet = pd.read_csv(sheet)
    sheet = sheet[sheet.columns[1:4]]

    sheet['i5_Tag'] = pd.Series([
        f"sample{index}_barcode_i5"
        for index in sheet.index.values
    ])
    sheet['i7_Tag'] = pd.Series([
        f"sample{index}_barcode_i7"
        for index in sheet.index.values
    ])

    config_dict['adapters'] = {"i7": adapter_i7, "i5": adapter_i5}

    tags_i5 = {row['i5_Tag']: row['i5_Barcode_Seq']
               for _, row in sheet.iterrows()}

    tags_i7 = {row['i7_Tag']: row['i7_Barcode_Seq']
               for _, row in sheet.iterrows()}

    config_dict['tag sequences'] = {**tags_i5, **tags_i7}

    config_dict['tag map'] = {row['Customer_Code']: f"{row['i5_Tag']},{row['i7_Tag']}"
                              for _, row in sheet.iterrows()}

    config_dict['names'] = {row['Customer_Code']: row['Customer_Code']
                            for _, row in sheet.iterrows()}
    return config_dict


def prepare_assembly_conf(output, samples):
    config_dict = dict()
    config_dict['samples'] = {sample: f'{output}/clean_fastq/{sample}/split-adapter-quality-trimmed/'
                              for sample in samples}
    return config_dict


def create_output(output):
    output = f'{os.getcwd()}/{output}'
    if os.path.isdir(output):
        raise IOError('Error: output directory already exists!\n'
                      'Rename or remove it before running UCEasy.')
    os.mkdir(output)


def get_samples_from_csv(sheet):
    sheet = pd.read_csv(sheet)
    samples = [row['Customer_Code'] for _, row in sheet.iterrows()]

    return samples


def render_taxon_set_conf(file, samples):
    string = '[all]'
    for sample in samples:
        string += f'\n{sample}'

    config = configparser.ConfigParser(allow_no_value=True)
    config.optionxform = str
    config.read_string(string)
    
    with open(file, 'w') as fl:
        config.write(fl, space_around_delimiters=False)

    return file
